program busmodel (input, output);
    const
	max_drives = 16;
	max_busses = 4;
	max_sources = 20;
    type
	drive_number = 0 .. max_drives;
	bus_number = 0 .. max_busses;
	source_number = 0 .. max_sources;
	trans_priority = 0 .. 16;
	state_of_completion = (idle, drive_wait, seeking, bus_wait, rotating,
				transferring, complete);
	length_of_transfer = 0 .. 65536;
	link = ^io_event;
	io_event = 
	    record
		fptr,bptr : link;
		time : real;
		source_n : source_number;
		drive_n : drive_number;
		bus_n : bus_number;
		priority : trans_priority;
		length : length_of_transfer;
		state : state_of_completion;
		time_of_qio : real;
		time_of_drive_grant : real;
		time_of_seek_complete : real;
		time_of_bus_grant : real;
		time_of_rotation_complete : real;
		time_of_transfer_complete : real
	    end;
	var
	    current_time : real;
	    current_event : link;
	    source : source_number;
	    drive : drive_number;
	    free_drive : drive_number;
	    bus : bus_number;
	    free_bus : bus_number;
	    seed : integer;
	    max_time : real;
	    print_records_flag : boolean;
	    utilization : real;
	    epoch : integer;

	    number_of_sources : source_number;
	    number_of_drives : drive_number;
	    number_of_busses : bus_number;

	    event_queue_base : link;

	    bus_busy : array [bus_number] of boolean;
	    bus_wait_queue_base : array [bus_number] of link;

	    drive_busy : array [drive_number] of boolean;
	    drive_wait_queue_base : array [drive_number] of link;

	    bus_serving_drive : array [drive_number] of bus_number;
	    max_seek_time : array [drive_number] of real;
	    max_latency_time : array [drive_number] of real;
	    transfer_rate : array [drive_number] of real;

	    drive_serving_source : array [source_number] of drive_number;
	    mean_qio_rate : array [source_number] of real;
	    mean_length : array [source_number] of real;
	    priority_of_source : array [source_number] of trans_priority;

	    mean_total_io_time : array [source_number] of real;
	    mean_drive_wait_time : array [source_number] of real;
	    mean_bus_wait_time : array [source_number] of real;
	    total_qios_processed : array [source_number] of integer;

	    time_drive_busy : array [drive_number] of real;
	    time_bus_busy : array [bus_number] of real;

    procedure place_in_io_event_queue (event : link);
	var
	   ev : link;
	begin
	    { link event record into chronological event queue }
	    ev := event_queue_base;
	    repeat
		ev := ev^.bptr
	    until (event^.time >= ev^.time) or (ev = event_queue_base);
	    event^.fptr := ev^.fptr;
	    event^.bptr := ev;
	    ev^.fptr^.bptr := event;
	    ev^.fptr := event
	end;
    procedure place_in_drive_wait_queue (event : link);
	var
	   ev : link;
	begin
	    event^.state := drive_wait;
	    { link event into decreasing priority ordered drive wait queue }
	    ev := drive_wait_queue_base [event^.drive_n];
	    repeat
		ev := ev^.bptr
	    until (event^.priority <= ev^.priority) 
		or (ev = drive_wait_queue_base [event^.drive_n]);
	    event^.fptr := ev^.fptr;
	    event^.bptr := ev;
	    ev^.fptr^.bptr := event;
	    ev^.fptr := event
	end; { place_in_drive_wait_queue }
    procedure place_in_bus_wait_queue (event : link);
	var
	   ev : link;
	begin
	    event^.state := bus_wait;
	    { link event into decreasing priority ordered bus wait queue }
	    ev := bus_wait_queue_base [event^.bus_n];
	    repeat
		ev := ev^.bptr
	    until (event^.priority <= ev^.priority)
		or (ev = bus_wait_queue_base [event^.bus_n]);
	    event^.fptr := ev^.fptr;
	    event^.bptr := ev;
	    ev^.fptr^.bptr := event;
	    ev^.fptr := event
	end; { place_in_drive_wait_queue }
    procedure extract_current_event_from (queue_base : link);
	begin
	    if queue_base^.fptr = queue_base then
		current_event := nil
	    else
		begin
		    current_event := queue_base^.fptr;
		    current_event^.fptr^.bptr := queue_base;
		    queue_base^.fptr := current_event^.fptr
		end;
	end; { extract_event }
    function mth$random (seed : integer) : real;
	extern;
    function uniform_rn : real;
	begin
	    uniform_rn := mth$random (seed)
	end;
    function exponential_rn : real;
	begin
	    exponential_rn := -ln( 1.0 - uniform_rn )
	end;
    procedure generate_new_event (source : source_number);
	var
	   new_event : link;
	   qio_delay_time : real;
	begin
	    { create a new event with the same attributes of the old event }
	    new (new_event);
	    with new_event^ do
		begin
		    source_n := source;
		    drive_n := drive_serving_source [source_n];
		    bus_n := bus_serving_drive [drive_n];
		    priority := priority_of_source [source_n];
		    { mean_length < 0 implies deterministic transfer length }
		    if mean_length [source_n] < 0.0 then
			length := trunc(-mean_length [source_n] / 512.0) + 1
		    else
			length := trunc(exponential_rn * 
					 mean_length [source_n] / 512.0) + 1;
		    length := length * 512;
		    state := idle;
		    qio_delay_time := exponential_rn / mean_qio_rate[source_n];
		    time_of_qio := current_time + qio_delay_time;
		    time := time_of_qio
		end; { with and do }
	    place_in_io_event_queue (new_event)
	end;
    procedure generate_seek;
	var
	    seek_delay : real;
	    rotation_delay : real;
	begin
	    { update io event record for seek }
	    with current_event^ do
		begin
		    state := seeking;
		    time_of_drive_grant := current_time;
		    drive_busy [drive_n] := true;
		    {
		The seek delay assumes independent uniformly distributed track
		seeks and quadratic head inertia.
		    }
		    seek_delay := sqrt( abs( uniform_rn - uniform_rn ) )
					* max_seek_time [drive_n];
		    {
		This seek operation is actually a SEARCH since the VAX massbus
		drivers do searches instead of seeks.  Thus the drive generates
		an interrupt when the track and sector requested are located by
		the read head.  The driver requests a sector 1/8 th revolution
		prior to the desired data to allow for the processing overhead
		required to allocate the massbus and request the transfer.
		    }
		    rotation_delay := uniform_rn * max_latency_time [drive_n];
		    time := current_time + seek_delay + rotation_delay;
		    time_of_seek_complete := time
		end; { with and do }
	    place_in_io_event_queue (current_event)
	end; { generate_seek }
    procedure generate_rotation;
	var
	    rotation_delay : real;
	begin
	    { update io event record for rotation }
	    with current_event^ do
		begin
		    state := rotating;
		    time_of_bus_grant := current_time;
		    bus_busy [bus_n] := true;
		    {
		The rotation delay at this point is deterministic because the
		seek (actually a search) produced a interupt at 1/8 of a 
		rotation before the desired sector. Thus the rotation delay is:
		(rot_time/8+time_of_seek_complete-current_time) mod rot_time
		Note that the MOD operator is not used here because it is 
		defined incorrectly for negative arguments by the creaters of 
		PASCAL.
		    }
		    rotation_delay := max_latency_time [drive_n] / 8.0 +
				time_of_seek_complete - current_time;
		    while rotation_delay < 0.0 do
			rotation_delay := rotation_delay + 
					  max_latency_time [drive_n];
		    time := current_time + rotation_delay;
		    time_of_rotation_complete := time
		end; { with and do }
	    place_in_io_event_queue (current_event)
	end; { generate_rotation }
    procedure generate_transfer;
	var
	    transfer_delay : real;
	begin
	    { update io event record for transfer }
	    with current_event^ do
		begin
		    state := transferring;
		    transfer_delay := length / transfer_rate[drive_n];
		    time := current_time + transfer_delay;
		    time_of_transfer_complete := time
		end; { with and do }
	    place_in_io_event_queue (current_event)
	end; { generate_transfer }
    procedure record_current_event_statistics;
	begin
	    with current_event^ do 
		begin
		    if print_records_flag then
		    writeln(source_n:3, drive_n:3, bus_n:3, priority:3, length,
			time_of_qio:8:3, time_of_drive_grant:8:3,
			time_of_seek_complete:8:3, time_of_bus_grant:8:3,
			time_of_rotation_complete:8:3,
			time_of_transfer_complete:8:3);
		    mean_total_io_time[source_n] :=
			mean_total_io_time[source_n] +
			time_of_transfer_complete - time_of_qio;
		    mean_drive_wait_time[source_n] :=
			mean_drive_wait_time[source_n] +
			time_of_drive_grant - time_of_qio;
		    mean_bus_wait_time[source_n] :=
			mean_bus_wait_time[source_n] +
			time_of_bus_grant - time_of_seek_complete;
		    total_qios_processed[source_n] :=
			total_qios_processed[source_n] + 1;
		    time_drive_busy[drive_n] := time_drive_busy[drive_n] +
			time_of_seek_complete - time_of_drive_grant +
			time_of_transfer_complete - time_of_bus_grant;
		    time_bus_busy[bus_n] := time_bus_busy[bus_n] +
			time_of_transfer_complete - time_of_bus_grant
		end; { with and do }
	end; { record_statistics }
    procedure create_double_linked_ring (var queue_base : link);
	begin
	    new (queue_base);
	    queue_base^.fptr := queue_base;
	    queue_base^.bptr := queue_base
	end;
    { Main program }
    begin
	writeln ('VAX 11/780 MASSbus Simulation');
	writeln;
	read (max_time, seed, print_records_flag,
		number_of_sources, number_of_drives, number_of_busses);
	writeln ('Time of simulation = ',max_time:7:2,
		';  Random seed = ', seed:7,
		';  Print IO records = ', print_records_flag:5);
	writeln	('Number of QIO sources =',number_of_sources:3,
		';  Number of disk drives =', number_of_drives:3,
		';  Number of busses =',  number_of_busses:3);
	for source := 1 to number_of_sources do
	    read (drive_serving_source [source],
		mean_qio_rate [source],		{ qio's / second }
		mean_length [source],		{ bytes / transfer }
		priority_of_source [source]);	{ qio priority }
	writeln;
	writeln ('QIO source  Drive used  QIOs/second  Mean bytes  Priority');
	for source := 1 to number_of_sources do
	    writeln (source:9,
		drive_serving_source [source]:10,
		mean_qio_rate [source]:14:2,
		mean_length [source]:13:1,
		priority_of_source [source]:8);
	for drive := 1 to number_of_drives do
	    read (bus_serving_drive [drive],
		max_seek_time [drive],		{ in seconds }
		max_latency_time [drive],	{ in seconds }
		transfer_rate [drive]);		{ in bytes / second }
	writeln;
	writeln ('Drive no.  Bus used  Max. seek time  Max. latency',
		'  Transfer rate');
	for drive := 1 to number_of_drives do
	    writeln (drive:8,
		bus_serving_drive [drive]:8,
		max_seek_time [drive]:15:6,
		max_latency_time [drive]:16:6,
		transfer_rate [drive]:16:1);
	create_double_linked_ring (event_queue_base);
	for bus := 1 to number_of_busses do
	    begin
		bus_busy [bus] := false;
		create_double_linked_ring (bus_wait_queue_base [bus])
	    end;
	for drive := 1 to number_of_drives do
	    begin
		drive_busy [drive] := false;
		create_double_linked_ring (drive_wait_queue_base [drive])
	    end;
	for source := 1 to number_of_sources do
	    generate_new_event (source);
	current_time := 0.0;
	for epoch := 1 to 2 do
	  begin
	  for source := 1 to number_of_sources do
	    begin
		mean_total_io_time [source] := 0.0;
		mean_drive_wait_time [source] := 0.0;
		mean_bus_wait_time [source] := 0.0;
		total_qios_processed [source] := 0
	    end;
	  for bus := 1 to number_of_busses do
	    time_bus_busy [bus] := 0.0;
	  for drive := 1 to number_of_drives do
	    time_drive_busy [drive] := 0.0;
	  repeat
	    { Extract next io_event from chronological queue }
	    extract_current_event_from (event_queue_base);
	    current_time := current_event^.time;
	    with current_event^ do
		case state of
		    idle :
			begin
			    if drive_busy [drive_n] then
				place_in_drive_wait_queue (current_event)
			    else
				generate_seek;
			    generate_new_event (source_n)
			end;
		    seeking :
			begin
			    if bus_busy [bus_n] then
				place_in_bus_wait_queue (current_event)
			    else
				generate_rotation
			end;
		    rotating :
			generate_transfer;
		    transferring :
			begin
			    state := complete;
			    drive_busy [drive_n] := false;
			    bus_busy [bus_n] := false;
			    record_current_event_statistics;
			    free_drive := drive_n;
			    free_bus := bus_n;
			    dispose (current_event);
			    extract_current_event_from
				(drive_wait_queue_base [free_drive]);
			    if current_event <> nil then generate_seek;
			    extract_current_event_from
				(bus_wait_queue_base[free_bus]);
			    if current_event <> nil then generate_rotation;
		    	end;
	    end; { with and case }
	  until current_time > epoch*max_time;
	  writeln;
	  writeln ('Run statistics for epoch ',epoch:3);
	  writeln ('QIO source  total QIOs  mean IO time  mean drive wait',
		'  mean bus wait');
	  for source := 1 to number_of_sources do
	    begin
		mean_total_io_time[source] := mean_total_io_time[source] /
						total_qios_processed[source];
		mean_drive_wait_time[source] := mean_drive_wait_time[source] /
						total_qios_processed[source];
		mean_bus_wait_time[source] := mean_bus_wait_time[source] /
						total_qios_processed[source];
		writeln (source:9, total_qios_processed[source]:11,
		    mean_total_io_time[source]:14:6,
		    mean_drive_wait_time[source]:14:6,
		    mean_bus_wait_time[source]:16:6)
	    end; { for and do }
	  writeln;
	  for drive := 1 to number_of_drives do
	    begin
		utilization := time_drive_busy[drive] / max_time * 100.0;
		writeln ('Percent utilization of drive ',drive:3,' is ',
		    utilization:5:2)
	    end;
	  writeln;
	  for bus := 1 to number_of_busses do
	    begin
		utilization := time_bus_busy[bus] / max_time * 100.0;
		writeln ('Percent utilization of bus ',bus:3,' is ',
		    utilization:5:2)
	    end;
	end; { for epoch }
    end. { main program }
